@@ -31,6 +31,14 @@ object HiddenApiHelper {
     @Volatile
     private var usersCacheTime: Long = 0
     private const val USERS_CACHE_TTL = 30_000L // 30 seconds
+
+    // Installed apps caching to avoid repeated expensive shell/API calls
+    // This is critical for performance - applyRules() calls getInstalledApplicationsAsUser() multiple times
+    @Volatile
+    private var installedAppsCache: MutableMap<Int, List<ApplicationInfo>> = mutableMapOf()
+    @Volatile
+    private var installedAppsCacheTime: Long = 0
+    private const val INSTALLED_APPS_CACHE_TTL = 5_000L // 5 seconds - short TTL to handle app installs/uninstalls
     
     /**
      * Data class representing a user profile
@@ -264,6 +272,9 @@ object HiddenApiHelper {
      * 2. Hidden API getInstalledApplicationsAsUser()
      * 3. Root shell: pm list packages --user X (fast, creates synthetic ApplicationInfo)
      *
+     * Results are cached for 5 seconds to avoid repeated expensive shell/API calls.
+     * This is critical for performance - applyRules() calls this multiple times per rule application.
+     *
      * Falls back to empty list if all methods fail.
      */
     fun getInstalledApplicationsAsUser(
@@ -273,11 +284,20 @@ object HiddenApiHelper {
     ): List<ApplicationInfo> {
         if (!initialized) initialize()
 
-        // For user 0, always use standard API (most reliable)
+        // For user 0, always use standard API (most reliable) - no caching needed as it's fast
         if (userId == 0) {
             return context.packageManager.getInstalledApplications(flags)
         }
 
+        // Check cache first for non-zero users (work profiles, clone profiles)
+        val now = System.currentTimeMillis()
+        if (now - installedAppsCacheTime < INSTALLED_APPS_CACHE_TTL) {
+            installedAppsCache[userId]?.let { cached ->
+                AppLogger.d(TAG, "📦 Returning cached ${cached.size} apps for user $userId")
+                return cached
+            }
+        }
+
         // Strategy 1: Try hidden API first
         if (hiddenApiAvailable) {
             try {
@@ -292,6 +312,7 @@ object HiddenApiHelper {
                 val apps = method.invoke(pm, flags, userId) as? List<ApplicationInfo>
                 if (!apps.isNullOrEmpty()) {
                     AppLogger.d(TAG, "✅ Found ${apps.size} apps for user $userId via hidden API")
+                    cacheInstalledApps(userId, apps)
                     return apps
                 }
             } catch (e: Exception) {
@@ -310,6 +331,7 @@ object HiddenApiHelper {
                 }
                 if (apps.isNotEmpty()) {
                     AppLogger.d(TAG, "✅ Found ${apps.size} apps for user $userId via shell (synthetic)")
+                    cacheInstalledApps(userId, apps)
                     return apps
                 }
             }
@@ -321,10 +343,31 @@ object HiddenApiHelper {
         return emptyList()
     }
 
+    /**
+     * Cache installed apps for a user and update the cache timestamp.
+     */
+    private fun cacheInstalledApps(userId: Int, apps: List<ApplicationInfo>) {
+        installedAppsCache[userId] = apps
+        installedAppsCacheTime = System.currentTimeMillis()
+    }
+
+    /**
+     * Clear the installed apps cache.
+     * Should be called when apps are installed/uninstalled or when a fresh list is needed.
+     */
+    fun clearInstalledAppsCache() {
+        installedAppsCache.clear()
+        installedAppsCacheTime = 0
+        AppLogger.d(TAG, "Cleared installed apps cache")
+    }
+
     /**
      * Create a synthetic ApplicationInfo for a work profile app.
      * Uses the personal profile's ApplicationInfo as a template and adjusts the UID.
      * This is much faster than calling pm dump for each package.
+     *
+     * IMPORTANT: The enabled state is queried from the work profile, not copied from personal profile,
+     * because an app can be enabled in personal profile but disabled in work profile.
      */
     private fun createSyntheticApplicationInfo(
         context: Context,
@@ -332,6 +375,9 @@ object HiddenApiHelper {
         userId: Int
     ): ApplicationInfo? {
         return try {
+            // Query the enabled state for this specific user (not from personal profile!)
+            val isEnabled = isPackageEnabledForUser(packageName, userId)
+
             // Try to get info from personal profile first (most apps are clones)
             val personalInfo = try {
                 context.packageManager.getApplicationInfo(packageName, 0)
@@ -346,6 +392,8 @@ object HiddenApiHelper {
                     // UID = userId * 100000 + appId
                     val appId = personalInfo.uid % 100000
                     this.uid = userId * 100000 + appId
+                    // Override enabled state with the actual state for this user profile
+                    this.enabled = isEnabled
                 }
             } else {
                 // App only exists in work profile, create minimal info
@@ -355,6 +403,7 @@ object HiddenApiHelper {
                     // Estimate UID - this may not be accurate but is good enough for display
                     this.uid = userId * 100000 + 10000 + packageName.hashCode().and(0xFFFF)
                     this.flags = 0
+                    this.enabled = isEnabled
                 }
             }
         } catch (e: Exception) {
@@ -398,6 +447,70 @@ object HiddenApiHelper {
         }
     }
 
+    /**
+     * Get the set of disabled packages for a specific user via shell.
+     * Uses `pm list packages -d --user $userId` to get disabled packages.
+     * Results are cached per user to avoid repeated shell calls.
+     */
+    private val disabledPackagesCache = mutableMapOf<Int, Set<String>>()
+
+    private fun getDisabledPackagesForUser(userId: Int): Set<String> {
+        // Return cached result if available
+        disabledPackagesCache[userId]?.let { return it }
+
+        return try {
+            val cachedShell = Shell.getCachedShell()
+            if (cachedShell == null || !cachedShell.isRoot) {
+                AppLogger.d(TAG, "No cached root shell available for disabled packages check (user $userId)")
+                return emptySet()
+            }
+
+            val outputList = mutableListOf<String>()
+            val result = cachedShell.newJob()
+                .add("pm list packages -d --user $userId")
+                .to(outputList)
+                .exec()
+
+            if (!result.isSuccess) {
+                AppLogger.d(TAG, "Shell pm list packages -d failed for user $userId: exit code ${result.code}")
+                return emptySet()
+            }
+
+            // Parse output: "package:com.example.app" -> "com.example.app"
+            val disabledSet = outputList
+                .filter { it.startsWith("package:") }
+                .map { it.removePrefix("package:").trim() }
+                .filter { it.isNotEmpty() }
+                .toSet()
+
+            // Cache the result
+            disabledPackagesCache[userId] = disabledSet
+            AppLogger.d(TAG, "Found ${disabledSet.size} disabled packages for user $userId")
+            disabledSet
+        } catch (e: Exception) {
+            AppLogger.d(TAG, "Shell pm list packages -d failed: ${e.message}")
+            emptySet()
+        }
+    }
+
+    /**
+     * Check if a package is enabled for a specific user.
+     * Uses the cached disabled packages set for performance.
+     */
+    private fun isPackageEnabledForUser(packageName: String, userId: Int): Boolean {
+        val disabledPackages = getDisabledPackagesForUser(userId)
+        return !disabledPackages.contains(packageName)
+    }
+
+    /**
+     * Clear the disabled packages cache.
+     * Should be called when package state might have changed.
+     */
+    fun clearDisabledPackagesCache() {
+        disabledPackagesCache.clear()
+        AppLogger.d(TAG, "Cleared disabled packages cache")
+    }
+
     /**
      * Get application info for a specific package and user.
      *
@@ -497,12 +610,16 @@ object HiddenApiHelper {
             // Determine if it's a system app
             val isSystem = flagsMatch?.groupValues?.get(1)?.contains("SYSTEM") == true
 
+            // Query the enabled state for this specific user
+            val isEnabled = isPackageEnabledForUser(packageName, userId)
+
             // Create ApplicationInfo
             ApplicationInfo().apply {
                 this.packageName = packageName
                 this.uid = absoluteUid
                 this.sourceDir = codePath ?: "/data/app/$packageName"
                 this.flags = if (isSystem) ApplicationInfo.FLAG_SYSTEM else 0
+                this.enabled = isEnabled
 
                 // Try to get the label from the personal profile version
                 try {
